/*************************************************************
Module		: peci driver module
*************************************************************/
/* this enables dbg prints */
#define DEBUG_ENABLE  1

#if DEBUG_ENABLE == 1
#define dbgprint         printk
#else
#define debgprint
#endif

/* Version Number of this module */
#define SAMPLE_MAJOR	1
#define SAMPLE_MINOR	0

#include <linux/module.h>
#include <linux/kernel.h>	/* printk() 		*/
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/ide.h>
#include <linux/delay.h>

#include "peci.h"

#define PECI_MAJOR	45
#define PECI_DEV_NAME   "PECI"

/*Control reg bits for domain 0(0-3) and domain 1(4-7)*/
#define DOMAIN_0_BASE		0
#define DOMAIN_1_BASE		4

/*PECI control ,status and data register*/
#define PECI_CTRL_REG           (*(volatile unsigned long*)0xB0201080)
#define PECI_STATUS_REG         (*(volatile unsigned long*)0xB0201084)
#define PECI_DATA_REG           (*(volatile unsigned long*)0xB0201088)

/*PECI Command execution time- roughly given based (there is nothing specified in the spec) */
#define PECI_INTERVAL   200

/*Write and read control register macros*/
#define WRITE_CTRL_REG(DATA)    PECI_CTRL_REG = (DATA)
#define READ_CTRL_REG(DATA)     (DATA) = PECI_CTRL_REG
#define CLEAR_CTRL_REG()        PECI_CTRL_REG = 0

/*Write and read status register macros*/
#define WRITE_STATUS_REG(DATA)    PECI_STATUS_REG = (DATA)
#define READ_STATUS_REG(DATA)     (DATA) = PECI_STATUS_REG
#define CLEAR_STATUS_REG()        PECI_STATUS_REG=0x0F

/*Write and read data register macros*/
#define READ_DATA_REG(DATA)     (DATA) = PECI_DATA_REG
#define CLEAR_DATA_REG()        PECI_DATA_REG = 0

/*PECI device errors return by Status registers (PECI Controller)*/
#define PECI_DEV_NO_ERROR       0x00
#define PECI_DEV_NEGO_FAILED    0x01
#define PECI_DEV_MISSING_NO_REP 0x02


/* Prototypes */
static int peci_open(struct inode * inode, struct file * filp);
static int peci_ioctl(struct inode *inode, struct file *file,
                     unsigned int cmd, unsigned long arg);
static int peci_init(void);
static int detect_cpu_presence(unsigned char cpu_index, unsigned char *presence, int domain);
static int scan_temperature(unsigned char cpu_index, unsigned char domain, unsigned short *ptemp, unsigned char *pstatus);

static struct file_operations peci_fops = {
	owner:		THIS_MODULE,
	read:		NULL,
	write:		NULL,
	ioctl:		peci_ioctl,
	open:		peci_open,
	release:	NULL,
};

/* semaphore to provide mutex */
struct semaphore g_sem;


/*driver initialization function*/
static int
init_peci_module(void)
{
	int 	err;

	printk(KERN_ALERT "PECI  Driver Version %d.%d\n",SAMPLE_MAJOR,SAMPLE_MINOR);
	printk(KERN_ALERT "Copyright (c) 2008 American Megatrends Inc.\n");

	/* ----  Register the character device ------------------- */
	if ((err = register_chrdev(PECI_MAJOR, PECI_DEV_NAME, &peci_fops)) < 0)
	{
		printk (KERN_ERR "failed to register device (err %d)\n", err);
		return -ENODEV;
	}

	/*Clear all the peci registers*/	
	peci_init();

	/* Initialize the semaphore */
	init_MUTEX(&g_sem);
	
	return 0;
}

/*driver exit function*/
static void
cleanup_peci_module(void)
{
        printk(KERN_ALERT "Unloading PECI Module ..............\n");
	
	/* uninitialize semaphore */
	init_MUTEX(&g_sem);

        /* unregister the device */
        unregister_chrdev(PECI_MAJOR, PECI_DEV_NAME);
	
        printk(KERN_ALERT "PECI module unloaded sucessfully.\n");
}
   

/* driver open function*/
static int peci_open(struct inode * inode, struct file * filp)
{
        if(filp->f_flags & 010000000)
        {
                return 0;
        }
        filp->f_op =&peci_fops;
	return  0;
}

/*do the driver initializations here*/
static int peci_init( void )
{
	/* clear all registers*/	
	CLEAR_CTRL_REG();
	CLEAR_STATUS_REG();
	CLEAR_DATA_REG();
	return 0;
}


/*peci ioctl handler*/
static int peci_ioctl(struct inode *inode, struct file *file,
                     unsigned int cmd, unsigned long arg)
{
	ioctl_peci_data_t 	pecidata;
	
	
	down(&g_sem);

	if ( copy_from_user(&pecidata, (void*)arg, sizeof(ioctl_peci_data_t)) )
	{
		up(&g_sem);
                return -EFAULT;
	}
	switch ( cmd )
	{
		case PECI_CPU_SCAN:
			pecidata.retvalue = detect_cpu_presence ( pecidata.cpu_index,&pecidata.status,pecidata.domain);	
		break;
		case PECI_GET_TEMP:
			pecidata.retvalue = scan_temperature ( pecidata.cpu_index, pecidata.domain, 
								&pecidata.temp, &pecidata.status );
		break;			
		default:
			dbgprint (" PECI:Invalid command\n");
			pecidata.retvalue = PECI_INVALID_CMD;
		break;
	}

	if ( copy_to_user((void*)arg,(void*)&pecidata, sizeof(ioctl_peci_data_t)) )
	{
		up(&g_sem);
		return -EFAULT;
	}
	up(&g_sem);
	return pecidata.retvalue;
}


/*
 * Detects CPU presence
 * input : cpu index, presence 
 * returns PECI_NO_ERROR on success , PECI_INTERNAL_ERROR and PECI_INVALID_PARAMS on failure 
 */
static int detect_cpu_presence ( unsigned char cpu_index, unsigned char *bpresence, int  domain )
{
	unsigned int temp 	 = 0;
	unsigned int ctrl_reg 	 = 0;
	unsigned int status_reg  = 0;
	unsigned int domain_base = 0;

	if ( cpu_index > MAX_CPU_COUNT )
		return PECI_INVALID_PARAMS;

	if ( 0 == domain )
	{
		domain_base = DOMAIN_0_BASE;
	} else if ( 1 == domain )
	{
		domain_base = DOMAIN_1_BASE;
	}else
	{
		dbgprint ("PECI: Invalid domain %d\n",domain);
		return PECI_INVALID_PARAMS;
	}

	temp = ( 1 << (cpu_index + domain_base) );
	CLEAR_STATUS_REG();
	CLEAR_CTRL_REG();
	WRITE_CTRL_REG( temp & 0x0F   );
	udelay(PECI_INTERVAL);
	READ_CTRL_REG (ctrl_reg);
	if ( ctrl_reg & (temp & 0x0F) )
	{
		dbgprint ("PECI: Busy  %d\n",(ctrl_reg&0x0F));
		return  PECI_INTERNAL_ERROR;
	}
	else
	{
		READ_STATUS_REG(status_reg);
		if ( (status_reg & 0x0F) & temp )
			*bpresence = 1;
		else
			*bpresence = 0;
		/*error status returned by peci wont applicable to presence detection*/
	}
	return PECI_NO_ERROR;
}

/*
 * Scan Temperature
 * input : cpu_index, domain, temp, status
 * returns PECI_NO_ERROR on success, PECI_INTERNAL_ERROR, PECI_CPU_NOT_FOUND and PECI_INVALID_PARAMS on error.
 */
static int scan_temperature ( unsigned char cpu_index, unsigned char domain, unsigned short *ptemp, unsigned char *pstatus )
{
	unsigned int temp =0;
	unsigned int ctrl_reg =0;
	unsigned char presence =0;
	int retvalue =0;
	
	retvalue =  detect_cpu_presence ( cpu_index, &presence, domain); 
	if ( PECI_NO_ERROR != retvalue )
	{
		return retvalue;	
	}
	
	if ( !presence )
	{
		*pstatus = 0;
		return PECI_CPU_NOT_FOUND;
	}	
	
	CLEAR_CTRL_REG();
	if ( 1 == domain )
	{
		temp = DOMAIN_1_BASE +  cpu_index;
	}
	else if ( 0 == domain )
	{
		temp = DOMAIN_0_BASE + cpu_index;
	}
	else
	{
		dbgprint("PECI: Invalid domain  \n");
		return PECI_INVALID_PARAMS;
	}
	if ( temp > 7 )
	{
		dbgprint("PECI: Invalid Processor Index\n");
		return PECI_INVALID_PARAMS;
	}
	if ( domain == 1 )
		printk (KERN_ALERT "ctrl_reg1 %x\n",(1<<temp));
	WRITE_CTRL_REG( 1<<temp);
	udelay(PECI_INTERVAL);
	READ_CTRL_REG (ctrl_reg);
	if ( domain == 1 )
                printk (KERN_ALERT "ctrl_reg1 %x\n",(1<<temp));
	if ( ctrl_reg & 0xFF )
	{
		dbgprint ("PECI: Busy \n");
		READ_STATUS_REG(*pstatus);
		return PECI_INTERNAL_ERROR;
	}
	else
	{
		READ_STATUS_REG(*pstatus);
		if ( *pstatus & 0x30 )
			return PECI_INTERNAL_ERROR;
		READ_DATA_REG(*ptemp);
	}

	return PECI_NO_ERROR;
}



/*--------------------------- Module Information Follows --------------------------*/
module_init (init_peci_module);
module_exit (cleanup_peci_module);

MODULE_AUTHOR("American Megatrends Inc");
MODULE_DESCRIPTION("This is PECI driver module.");
MODULE_LICENSE("GPL");

