/*
 * $Id$
 *
 * Copyright 2008, Network Appliance Inc.
 * Author: Jason McMullan <mcmullan@netapp.com>
 *
 */

#include <common.h>
#include <i2c.h>
#include <asm/io.h>

#include "vsc452/vsc452_i2c.h"

#define I2C_TIMEOUT	200	/* 200ms */

#define WAIT_UNTIL(condition, timeout_ms) \
	do { \
		int time_left = timeout_ms * 1000; \
		while (time_left > 0 && !(condition)) { \
			udelay((time_left > 100) ? 100 : time_left); \
			time_left -= (time_left > 100)  ? 100 : time_left; \
		}; \
	} while (0); if (!(condition))

int i2c_bus;

static struct vsc452_i2c_bus {
	int speed;	/* offset into the speed map */
	uint32_t base;	/* Physical address */
} i2c[VSC452_I2C_INTERFACES];

static const struct {
	unsigned int speed;
	unsigned int select;
} i2c_speed_map[] = {
	{ .speed =   78000, .select = SID_CMS_CLK_78KHZ },
	{ .speed =   98000, .select = SID_CMS_CLK_98KHZ },
	{ .speed =  284000, .select = SID_CMS_CLK_284KHZ },
	{ .speed =  340000, .select = SID_CMS_CLK_340KHZ },
	{ .speed =  347000, .select = SID_CMS_CLK_347KHZ },
	{ .speed =  417000, .select = SID_CMS_CLK_417KHZ },
	{ .speed =  852000, .select = SID_CMS_CLK_852KHZ },
	{ .speed = 1042000, .select = SID_CMS_CLK_1042KHZ },
};

static inline uint32_t read32(uint32_t base)
{
	return *(volatile uint32_t *)(KSEG1ADDR(base));
}

static inline void write32(uint32_t base, uint32_t val)
{
	*(volatile uint32_t *)(KSEG1ADDR(base)) = val;
}

static inline uint32_t iread(int reg)
{
	return read32(i2c[i2c_bus].base + reg);
}

static inline void iwrite(int reg, uint32_t val)
{
	write32(i2c[i2c_bus].base + reg, val);
}
int i2c_set_bus_speed (unsigned int speed)
{
        int i;

        for (i = 0; i < ARRAY_SIZE(i2c_speed_map); i++) {
                if (speed <= i2c_speed_map[i].speed) {
                        break;
                }
        }

        /* Too fast? Then select the fastest possible.
	 */
        if (i == ARRAY_SIZE(i2c_speed_map)) {
                i--;
        }

        i2c[i2c_bus].speed = i;
        /* Set the speed, do not snoop the bus */
        iwrite(SIC, SIC_PIN | SIC_ES_CMS);
        iwrite(SID, i2c_speed_map[i2c[i2c_bus].speed].select);

        /* Return to data i/o mode */
        iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_ACK);

        return 0;
}

int i2c_set_bus_num (unsigned int bus)
{
	int i;

	if (bus >= VSC452_I2C_INTERFACES) {
		return -1;
	}

	i2c_bus = bus;
	i2c[i2c_bus].base = I2C_BASE(i2c_bus);

	/* Reset the controller */
	iwrite(SIBC, SIBC_SISR);
	iwrite(SIBC, SIBC_SCC | SIBC_SDC);

	/* Wiggle out any data on the line */
	if ((iread(SIBS) & SIBS_SDS) == 0) {
		for (i = 0; i < 8 * 16; i++) {
			iwrite(SIBC,        0 | SIBC_SDC);
			udelay(100);
			iwrite(SIBC, SIBC_SCC | SIBC_SDC);
			udelay(100);
		}
	}

	/* Master only */
	iwrite(SIC, SIC_PIN | SIC_ES_SCA);
	iwrite(SID, SID_SCA_ADDR(0x7F));

	/* Select the bus speed */
	i2c_set_bus_speed(i2c[i2c_bus].speed);

	return 0;
}

unsigned int i2c_get_bus_num (void)
{
	return i2c_bus;
}

unsigned int i2c_get_bus_speed (void )
{
	return i2c_speed_map[i2c[i2c_bus].speed].speed;
}


#define I2C_STAT(tmp) \
	do { \
		debug("Status: 0x%08x (", tmp); \
		if ((tmp & SIC_PIN) == SIC_PIN) debug("PIN "); \
		if ((tmp & SIC_SSN) == SIC_SSN) debug("SSN "); \
		if ((tmp & SIC_STS) == SIC_STS) debug("STS "); \
		if ((tmp & SIC_BER) == SIC_BER) debug("BER "); \
		if ((tmp & SIC_LRB) == SIC_LRB) debug("LRB "); \
		if ((tmp & SIC_AAS) == SIC_AAS) debug("AAS "); \
		if ((tmp & SIC_LAB) == SIC_LAB) debug("LAB "); \
		if ((tmp & SIC_BBN) == SIC_BBN) debug("BBN "); \
		debug(")\n"); \
	} while (0)

#define I2C_ABORT() do { \
	i2c_set_bus_num(i2c_bus); \
} while (0)


int i2c_probe(uchar chip)
{
	uchar c[2];

	return i2c_read(chip, 0, 0, &c[0], 2);
}

void i2c_init (int speed,int slaveaddr)
{
	// 	i2c_set_bus_speed (speed);
	i2c_set_bus_num (CFG_CONTROLLER_BUS_NO);
}

int i2c_read(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	uchar tmp;
	int i;
	
	if (len < 1 || alen < 0) {
		return -1;
	}

	/* Wait for bus to stop */
	WAIT_UNTIL((iread(SIC) & SIC_BBN) == SIC_BBN, I2C_TIMEOUT) {
		debug("I2C bus timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		return -1;
	}

	/* Send start */
	iwrite(SID, SID_DATA((chip << 1) | ((alen == 0) ? 1 : 0)));
	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STA | SIC_ACK);

	/* Wait for NAK */
	WAIT_UNTIL(((tmp = iread(SIC)) & SIC_PIN) == 0, I2C_TIMEOUT) {
		I2C_STAT(tmp);
		debug("I2C address timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		return -1;
	}

	if (tmp & SIC_LRB) {
		debug("I2C slave NACK, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);
		return -1;
	}

	/* Send address */
	for (i = 0; i < alen; i++) {
		I2C_STAT(tmp);
		if ((tmp & SIC_LRB) != 0) {
			debug("I2C slave missing, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			return -1;
		}

		iwrite(SID, SID_DATA(addr >> ((alen - i - 1) * 8)));
		WAIT_UNTIL(((tmp = iread(SIC)) & SIC_PIN) == 0, I2C_TIMEOUT) {
			I2C_STAT(tmp);
			debug("I2C address timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			return -1;
		}
	}

	if (alen > 0) {
		/* Send chip id */
		iwrite(SID, SID_DATA((chip << 1) | 1));
		/* Send repeated start */
		iwrite(SIC, SIC_ES_DATA | SIC_STA | SIC_ACK);
	}


	for (i = 0; i < len; i++) {
		buffer[i] = 0x80 + i;
	}
	for (i = -1; i < (len - 1); i++) {
		WAIT_UNTIL((iread(SIC) & SIC_PIN) == 0, I2C_TIMEOUT) {
			debug("I2C read timeout, bus = %d, chip = 0x%02x, offset %d\n", i2c_bus, chip, i);
			I2C_ABORT();
			return -1;
		}

		if ((iread(SIC) & SIC_LRB) != 0) {
			/* Error! */
			iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);
			debug("I2C read error, bus = %d, chip = 0x%02x, offset %d\n", i2c_bus, chip, i);
			tmp = iread(SID);	/* Throw-away data */
			I2C_ABORT();
			return -1;
		}

		tmp = SID_DATA(iread(SID));
		if (i >= 0) {
			buffer[i] = tmp;
		}
	}

	iwrite(SIC, SIC_ES_DATA);
	tmp = SID_DATA(iread(SID));
	if (i >= 0) {
		buffer[i] = tmp;
	}

	WAIT_UNTIL((iread(SIC) & SIC_PIN) == 0, I2C_TIMEOUT) {
		debug("I2C finish error, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		return -1;
	}

	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);
	tmp = SID_DATA(iread(SID));
	buffer[i] = tmp;

	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_ACK);
	return 0;
}

int i2c_write(uchar chip, uint addr, int alen, uchar * buffer, int len)
{
	int i;
	uchar tmp;

	/* Wait for bus to stop */
	WAIT_UNTIL((iread(SIC) & SIC_BBN) == SIC_BBN, I2C_TIMEOUT) {
		debug("I2C bus timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		return -1;
	}

	/* Send start */
	iwrite(SID, SID_DATA((chip << 1) | 0));
	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STA | SIC_ACK);

	/* Wait for NAK */
	WAIT_UNTIL(((tmp = iread(SIC)) & SIC_PIN) == 0, I2C_TIMEOUT) {
		I2C_STAT(tmp);
		debug("I2C NAK timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		return -1;
	}

	if (tmp & SIC_LRB) {
		debug("I2C slave NACK, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);
		return -1;
	}

	/* Send address */
	for (i = 0; i < alen; i++) {
		iwrite(SID, SID_DATA(addr >> ((alen - i - 1) * 8)));

		WAIT_UNTIL(((tmp = iread(SIC)) & SIC_PIN) == 0, I2C_TIMEOUT) {
			I2C_STAT(tmp);
			debug("I2C address timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			return -1;
		}

		if ((iread(SIC) & SIC_LRB) != 0) {
			debug("I2C slave missing, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			return -1;
		}
	}

	/* Send data */
	for (i = 0; i < len; i++) {
		iwrite(SID, SID_DATA(buffer[i]));

		WAIT_UNTIL((iread(SIC) & SIC_PIN) == 0, I2C_TIMEOUT) {
			debug("I2C address timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			return -1;
		}

		if ((iread(SIC) & SIC_LRB) != 0) {
			debug("I2C slave missing, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			return -1;
		}
	}

	/* Send stop */
	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);


	return 0;
}

